/*
 * external_call.c
 *
 * This module supports external calls by execve.
 *
 * One call is with respect to initialisation of the operating system.
 * One call is to obtain the hardware identity of the product by bootblocktool.
 *
 * (C) Copyright 2002-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "msg.h"
#include "flash.h"

#define D(x)

/* Error messages. */
static const char error_pipe[] = "pipe";
static const char error_dup[] = "dup";
static const char error_close[] = "close";
static const char error_execve[] = "execve";
static const char error_waitpid[] = "waitpid";
static const char error_access[] = "access";
static const char error_fcntl[] = "fcntl";
static const char error_fclose[] = "fclose";
static const char error_read[] = "read";
static const char error_write[] = "write";
static const char error_getrlimit[] = "getrlimit";

/* External calls. */
static char prescript_exec[] = "/bin/preflashupgrade";
static char *bbt_argv[] = {"/bin/bootblocktool", "-nocgi", "-x", "HWID", NULL};

/*
 * Return 0 if script is successfully executed with arguments argv or -1.
 * Output from script is forwarded to statfd.
 */
static int run_script(const char *script, char *argv[], int statfd);

/* Return 0 if the operating system is ready for flash upgrade or -1. */
int run_preparation_script(char *target_string, char *protocol_string,
			   int statfd);

/* Return a string containing the HWID of a unit or NULL on failure. */
char *get_hwid_from_bootblock(void);

/*
 * Return 0 if script is successfully executed with arguments argv or -1.
 * Output from script is forwarded to statfd.
 */
static int
run_script(const char *script, char *argv[], int statfd)
{
	int fd[2];
	char buffer[81]; /* room for one line and a line feed character */
	pid_t pid;
	int status;
	ssize_t read_byte;

	D(msg(__FUNCTION__));
	D(msg(script));
	D({int i = 0; while (argv[i] != NULL) msg(argv[i++]);})

	/* Do a basic check on the script. */
	if (access(script, X_OK) < 0) {
		errmsg_errno(error_access, errno);
		return -1;
	}

	/* pipe[0] is for reading, pipe[1] is for writing */
	if (pipe(fd) < 0) {
		errmsg_errno(error_pipe, errno);
		return -1;
	}

	if ((pid = fork()) < 0) {
		return -1;
	} else if (pid == 0) {
		struct rlimit rlim;
		rlim_t fdcount;

		close(fd[0]);

		/* child. */
		if (dup2(fd[1], STDOUT_FILENO) < 0) {
			errmsg_errno(error_dup, errno);
			_exit(EXIT_FAILURE);
		}
		if (close(fd[1]) < 0) {
			errmsg_errno(error_close, errno);
			_exit(EXIT_FAILURE);
		}

		/* Close all fds except stdin, stdout and stderr. */
		if (getrlimit(RLIMIT_NOFILE, &rlim) < 0) {
			errmsg_errno(error_getrlimit, errno);
			_exit(EXIT_FAILURE);
		}
		for (fdcount = STDERR_FILENO + 1; fdcount < rlim.rlim_cur;
		     fdcount++) {
			if ((close(fdcount) < 0) && (errno != EBADF)) {
				errmsg_errno(error_close, errno);
				_exit(EXIT_FAILURE);
			}
		}

		if (execv(script, argv) < 0) {
			errmsg_errno(error_execve, errno);
			_exit(EXIT_FAILURE);
		}
		_exit(EXIT_FAILURE);
	}
	close(fd[1]);

	/* Forward stdout from script to statfd. */
	while (((read_byte = read(fd[0], buffer, sizeof(buffer) - 1)) > 0) ||
		((read_byte == -1) && (errno == EAGAIN))) {
		if (read_byte > 0) {
			if (write(statfd, buffer, read_byte) < 0) {
				errmsg_errno(error_write, errno);
			}
		}
	}
	close(fd[0]);

	if (read_byte < 0) {
		errmsg_errno(error_read, errno);
	}

	/* Wait for script to finish. */
	if (waitpid(pid, &status, 0) < 0) {
		errmsg_errno(error_waitpid, errno);
		return -1;
	}

	/* Forward messages. */
	if (WIFEXITED(status) && WEXITSTATUS(status)) {
		/* Child exited normally, but the exit status was != 0. */
		D(msg("run_script: normal exit, status != 0"));
		return -1;
	} else if (!WIFEXITED(status)) {
		/* Child did not exit normally. */
		D(msg("run_script: not normal exit"));
		return -1;
	} else {
		D(msg("run_script: normal exit"));
		return 0;
	}
}

/*
 * Return 0 if the operating system is ready for flash upgrade or -1.
 * Output from script is forwarded to statfd.
 */
int
run_preparation_script(char *target_string, char *protocol_string, int statfd)
{
	char *argv[4];

	argv[0] = prescript_exec;
	argv[1] = target_string;
	argv[2] = protocol_string;
	argv[3] = NULL;
	return run_script(prescript_exec, argv, statfd);
}

/* Return a string containing the HWID of a unit or NULL on failure. */
char *
get_hwid_from_bootblock(void)
{
	int fd[2];
	pid_t pid;
	int status;
	char *nl;
	static char line[HWID_LIST_LEN + 1];

	D(msg(__FUNCTION__));

	memset(line, 0, sizeof(line));

	if (pipe(fd) < 0) {
		errmsg_errno(error_pipe, errno);
		return NULL;
	}

	if ((pid = fork()) < 0) {
		return NULL;
	} else if (pid == 0) {
		/* child. */
		if (fclose(stdout) < 0) {
			D(perror(error_fclose));
			_exit(EXIT_FAILURE);
		}
		if (dup2(fd[1], STDOUT_FILENO) < 0) {
			D(perror(error_dup));
			_exit(EXIT_FAILURE);
		}
		if (close(fd[1]) < 0) {
			D(perror(error_close));
			_exit(EXIT_FAILURE);
		}

		if (execv(bbt_argv[0], bbt_argv) < 0) {
			D(perror(error_execve));
			_exit(EXIT_FAILURE);
		}
		_exit(EXIT_FAILURE);
	}
	/* parent. */
	if (waitpid(pid, &status, 0) < 0) {
		errmsg_errno(error_waitpid, errno);
		return NULL;
	}
	if (fcntl(fd[0], F_SETFL, O_NONBLOCK) < 0) {
		errmsg_errno(error_fcntl, errno);
		return NULL;
	}
	(void)read(fd[0], line, sizeof(line) - 1);
	if ((nl = strchr(line, '\n'))) {
		*nl = '\0';
	} else {
		line[HWID_LIST_LEN] = '\0';
	}

	/* Forward messages. */
	if (WIFEXITED(status) && WEXITSTATUS(status)) {
		/* Child exited normally, but the exit status was != 0. */
		D(msg("bootblocktool: normal exit, status != 0"));
		return NULL;
	} else if (!WIFEXITED(status)) {
		/* Child did not exit normally. */
		D(msg("bootblocktool: not normal exit"));
		return NULL;
	} else {
		D(msg("bootblocktool: normal exit"));
		if (!strlen(line)) {
			return NULL;
		}
		return line;
	}
}
