/*
 * flash.c
 *
 * This module supports flash memory operations in upgrades.
 *
 * (C) Copyright 2002-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <sys/mount.h>
#include <linux/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/version.h>
#include <sys/mman.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION (2, 6, 8)
#if LINUX_VERSION_CODE < KERNEL_VERSION (2, 6, 19)
#include <linux/compiler.h>
#endif
#include <mtd/mtd-user.h>
#else
#include <linux/mtd/mtd.h>
#endif
#include "msg.h"
#include "external_call.h"
#include "flash.h"

/* define for debug printouts */
#define D(x)

#ifdef NANDFLASH_SUPPORT
#if !defined(MEMGETBADBLOCK) || !defined(MEMSETBADBLOCK)
#error Misconfiguration - configured for NAND flash, but not supported by kernel!
#endif
#endif

enum partition_type {
	PARTITION_TYPE_RESCUE = 0x0000,
	PARTITION_TYPE_PARAM  = 0x0001,
	PARTITION_TYPE_KERNEL = 0x0002,
	PARTITION_TYPE_JFFS   = 0x0003,
	PARTITION_TYPE_UNDEF  = 0x0FFF
};

/* commands for next_iov() function */
#define CMD_CURRENT_IOV 0
#define CMD_NEXT_IOV    1

/*
 * FIXME The checksum calculation of the image requires a non-zero length
 * which is unknown at the time of the calculation.
 */
#define UNKNOWN_LEN 1

/*
 * The format is taken from the rescue code definition. There will most
 * likely be three partitions since the format is difficult to change.
 *
 * The partitiontable is designed so that it is transparent to
 * code execution - it has a relative branch opcode in the
 * beginning that jumps over it. each entry contains extra
 * data so we can add stuff later.
 *
 * Partition table format example (for CRIS):
 *
 *     Code transparency:
 *
 *     2 bytes    [opcode 'nop']
 *     2 bytes    [opcode 'di']
 *     4 bytes    [opcode 'ba <offset>', 8-bit or 16-bit version]
 *     2 bytes    [opcode 'nop', delay slot]
 *
 *     Table validation (at +10):
 *
 *     2 bytes    [magic/version word for partitiontable - 0xef, 0xbe]
 *     2 bytes    [length of all entries plus the end marker]
 *     4 bytes    [checksum for the partitiontable itself]
 *
 *     Entries, each with the following format, last has offset -1:
 *
 *        4 bytes    [offset in bytes, from start of partition table]
 *        4 bytes    [length in bytes of partition]
 *        4 bytes    [checksum, simple longword sum]
 *        2 bytes    [partition type]
 *        2 bytes    [flags, only bit 0 used, ro/rw = 1/0]
 *        16 bytes   [reserved for future use]
 *
 *     End marker
 *
 *        4 bytes    [-1]
 *
 *       10 bytes    [0, padding]
 *
 * Bit 0 in flags signifies RW or RO. The rescue code only bothers
 * to check the checksum for RO partitions, since the others will
 * change its data without updating the checksums. A 1 in bit 0
 * means RO, 0 means RW. That way, it is possible to set a partition
 * in RO mode initially, and later mark it as RW, since you can always
 * write 0's to the flash.
 */

/*
   Man, this should really be rewritten for NAND support, instead of
   patching it rather brutally on top of older NOR code. Turned out
   hairier than I would have thought. Alas.
*/

#define PTABLE_MAGIC 0xBEEF
#define MAX_MAGIC_OFFSET 100

struct cur_iov {
	unsigned block_nr_in_i_partition; /* current block */
	unsigned block_nr_in_f_partition; /* current block */
	unsigned blocks_in_i_partition;   /* total nr of blocks in partition */
	unsigned blocks_in_f_partition;   /* total nr of blocks in partition */
	unsigned partition_nr;
	unsigned iov_nr;
	struct partition_layout *next;    /* next partition */
	unsigned skip;
	loff_t flash_addr;
	loff_t part_addr;

};


struct pt_head {
	unsigned short magic;   /* Magic BEEF. */
	unsigned short size;    /* Length of ptable block excluding header. */
	unsigned int checksum;  /* Check sum of memory. */
};

struct pt_entry {
	unsigned int offset;    /* Offset relative the sector of the ptable. */
	unsigned int size;      /* Size of the partition. */
	unsigned int checksum;  /* The check sum of the partition. */
	unsigned short type;    /* Refer to enum partition_type. */
	unsigned short flags;   /* Bit 0: read-only = 1, read-write = 0. */
	unsigned int reserved0; /* Reserved for future use. */
	unsigned int reserved1; /* Reserved for future use. */
	unsigned int reserved2; /* Reserved for future use. */
	unsigned int reserved3; /* Reserved for future use. */
};

/* Device names of the flash memory. */
static const char dev_name_rescue[] = "/dev/part/rescue";
static const char dev_name_kernel[] = "/dev/part/kernel";

/* defined in flash.h */
struct partition_layout *flash_layout;

/* four partitions */
struct partition_layout flash_layout_4[] = {
	{ 0, 0, 0, dev_name_rescue,    0, 0, NULL, 0, -1, 0, NULL },
	{ 0, 0, 0, dev_name_kernel,    0, 0, NULL, 0, -1, 0, NULL },
	{ 0, 0, 0, "/dev/part/rootfs", 0, 0, NULL, 0, -1, 0, NULL },
	{ 0, 0, 0, "/dev/part/rwfs",   0, 0, NULL, 0, -1, 0, NULL },
	{ 0, 0, 0, NULL,               0, 0, NULL, 0, -1, 0, NULL }
};

/* three partitions */
struct partition_layout flash_layout_3[] = {
	{ 0, 0, 0, dev_name_rescue,    0, 0, NULL, 0, -1, 0, NULL },
	{ 0, 0, 0, dev_name_kernel,    0, 0, NULL, 0, -1, 0, NULL },
	{ 0, 0, 0, "/dev/part/rwfs",   0, 0, NULL, 0, -1, 0, NULL },
	{ 0, 0, 0, NULL,               0, 0, NULL, 0, -1, 0, NULL }
};

/* Messages sent to the client. */
static const char error_eop[] = "Error: out of memory in partition";
static const char error_open[] = "Error: open";
static const char error_close[] = "Error: close";
static const char error_read[] = "Error: read";
static const char error_writev[] = "Error: writev";
static const char error_ioctl[] = "Error ioctl";
static const char error_lseek[] = "Error lseek";
static const char error_malloc[] = "Error: malloc";
static const char error_setrlimit[] = "Error: setrlimit";
static const char error_mlockall[] = "Error: mlockall";
static const char error_modulo[] =
	"Error: image size is not multiple of memory device blocksize";
static const char error_erase[] = "Failed to erase sector!";
static const char error_checksum_wrong[] = "Checksum of image is wrong!";
static const char error_hwid[] =
	"Wrong HWID! The uploaded firmware cannot run on this hardware.";
static const char error_find_ptable[] = "Could not find partition table!";
static const char error_compare_ptable[] =
	"Partition table for image differs from existing firmware.";
static const char error_verify[] = "Verification of flash memory failed!";
static const char error_paging[] = "Failed to disable paging!";
static const char info_no_hwid[] = "No HWID in bootblock, continuing anyway.";
static const char info_erasing[] = "Erasing old file system ...";
static const char info_loading[] = "Loading new file system ...";

/* external function prototypes */

/* Return 0 if the device for the flash memory partition was opened (O_RDWR)
   or return -1. */
extern int open_flash_device(struct partition_layout *partition);

/* Return 0 if paging is disabled for all pages mapped into the address space
   of the calling process, return -1 on failure. */
extern int disable_paging(void);

/* Return 0 if a partition table is found in the supplied buffer,
   which should be a pointer to the beginning of the image kernel
   partition, or return -1. The size and type of each partition are
   stored in layout. */
extern int read_image_partition_table(unsigned char *buf);

/* Return 0 if the HWID and checksum of the image are valid, return -1 if
   either or both are invalid. */
extern int verify_hwid_checksum(char *tail_buf, char *fwinfo_hwid);

/* Free the dynamically allocated memory used for image file storage. */
extern void free_iov_buffers(void);

/* Return blocksize/eraseize of partition, or -1 on fail */
extern int get_block_size(struct partition_layout *partition);

/* Return 0 if it is possible to allocate the buffers defined in
   partition_layout, return -1 if malloc fails. The partition_layout
   contains an initialized buffer pointer area with allocated memory
   if 0 is returned. */
extern int init_iov_buffers_p(struct partition_layout *layout);

/* Return 0 if a partition table is found in the flash memory kernel partition
   or return -1. The size and type of each partition are stored in layout. */
extern int read_flash_partition_table(void);

/* Return 0 if the image and flash partition tables are of equal lengths and
   types or return -1. */
extern int compare_partition_tables(void);

/* Close open file descriptors. */
extern void close_flash_devices(struct partition_layout *partition[],
                                enum flash_protocol protocol);

/* Return 0 if a flash memory partition can be erased, programmed and verified
   against the image data programed by or return -1. */
extern int update_flash(struct partition_layout *partition,
                        enum flash_protocol protocol);

/* return nr of partitions in flash, -1 on fail */
extern int get_nr_of_partitions();

/* Set length of an image partition in the flash_layout. */
extern void set_length(unsigned partition_nr, unsigned long len);

/* Return length of a flash partition from the flash_layout. */
extern unsigned long get_length(unsigned partition_nr);

/* static function prototypes */

/* Return 0 if the partition can be erased, return -1 on failure. Erasure is
   attempted 10 times for a sector. */
static int erase_flash(struct partition_layout *partition,
                       enum flash_protocol protocol);

/* Return 0 if the data from the image stored in partition can be written
   to flash memory, return -1 if writev fails. */
static int program_flash(struct partition_layout *partition,
                         enum flash_protocol protocol);

/* Return 0 if the content of flash memory equals that of the image written to
   flash memory, return -1 on failure or on differences. */
static int verify_flash(struct partition_layout *partition);

/* Get mtd_info. Return -1 on fail. */
static int get_mtd_info(int fd, mtd_info_t *mtd_info);

/* Returns the simple minded check sum of image partitions buffered in the
   flash_layout and the string hwid_image. */
static unsigned long int calc_checksum(struct partition_layout *layout,
                                       char *hwid_image);

/* Return 0 if the rescue sector length can be read in flash memory or -1. */
static int read_rescue_sector_length(unsigned int *length);

/* Print all percentages from last percentage if percent differs from
   last percent. Counter is reset and no percentage printed if called
   with 0 percent. I.e not reentrant safe. */
static void print_percent(int percent, enum flash_protocol protocol);

/* Iterator for image iovectors. Return pointer to next or current iov
   from image, depending on cmd, or NULL if no next iov. If valid iov,
   fills in info in cur_iov struct. Starts in KERNEL_PARTITION. Not
   reentrant safe. */
static struct iovec *next_iov(unsigned cmd, struct cur_iov *cur_iov);

static int skip1(struct cur_iov *cur_iov, struct iovec *iov,
                 struct partition_layout *partition);

/************************************************************************/

int
get_nr_of_partitions()
{
	int fd;
	mtd_info_t mtd_info;

	D(msg(__FUNCTION__));

	/* Open the rescue partition in flash memory. */
	if ((fd = open(dev_name_rescue, O_RDONLY)) < 0) {
		errmsg_errno(error_open, errno);
		return -1;
	}

	if (get_mtd_info(fd, &mtd_info) < 0) {
		errmsg_errno(error_ioctl, errno);
		return -1;
	}

	close(fd);

	if (mtd_info.type == MTD_NORFLASH) {
		flash_layout = flash_layout_3;
		D(msg("3 partitions\n"));
		return 3;
	} else {
		flash_layout = flash_layout_4;
		D(msg("4 partitions\n"));
		return 4;
	}
}

/* Return blocksize/eraseize of partition, or -1 on fail */
int
get_block_size(struct partition_layout *partition)
{
	mtd_info_t mtd_info;

	D(msg(__FUNCTION__));

	if (open_flash_device(partition) < 0) {
		msg(error_open);
		return -1;
	}

	if (get_mtd_info(partition->fd, &mtd_info) < 0) {
		errmsg_errno(error_ioctl, errno);
		return -1;
	}

	D(msg_int("erasesize", mtd_info.erasesize));

	partition->erasesize = mtd_info.erasesize;

	return partition->erasesize;
}


/* Set length of an image partition in the flash_layout. */
void
set_length(unsigned partition_nr, unsigned long len)
{
	D(msg_int("Setting partition length to", len));

	flash_layout[partition_nr].image_len = len;
}

/* Return length of a flash partition from the flash_layout. */
unsigned long
get_length(unsigned partition_nr)
{
	unsigned long len;

	len = flash_layout[partition_nr].flash_len;
	return len;
}

/* Return 0 if paging is disabled for all pages mapped into the address space
   of the calling process, return -1 on failure. */
int
disable_paging(void)
{
	struct rlimit value;

	D(msg(__FUNCTION__));

	value.rlim_max = RLIM_INFINITY;
	value.rlim_cur = RLIM_INFINITY;

	if (setrlimit(RLIMIT_MEMLOCK, &value) < 0) {
		errmsg_errno(error_setrlimit, errno);
		goto panic;
	}
	if (mlockall(MCL_CURRENT | MCL_FUTURE) < 0) {
		errmsg_errno(error_mlockall, errno);
		goto panic;
	}
	return 0;

panic:
	msg(error_paging);
	return -1;
}

/* Returns the simple minded check sum of image partitions buffered in the
   flash_layout and the string hwid_image. */
static unsigned long int
calc_checksum(struct partition_layout *layout, char *hwid_image)
{
	unsigned int sum;
	struct partition_layout *partition;
	struct iovec *iov;
	int i;
	int j;

	D(msg(__FUNCTION__));

	sum = 0;
	for (partition = layout; partition->image_len; partition++) {
		iov = partition->iov;
		for (i = 0; i < partition->iov_length; i++) {
			for (j = 0; j < iov[i].iov_len; j++)
				sum += *(unsigned char *)(iov[i].iov_base + j);
		}
		D(msg_int(partition->device_name, sum));
	}

	for (i = 0; i < HWID_LIST_LEN; i++) {
		sum += ((unsigned char *)hwid_image)[i];
	}
	D(msg_int("Calculated checksum", sum));

	return sum;
}

/* If tail_buf, return 0 if the HWID and checksum of the image are
   valid, return -1 if either or both are invalid. If !tail_buf but
   hwid_info, just compare hwid from bootblock with the hwid string in
   the hwid_fwinfo paramater, returns 0 on equal, otherwise -1. */
int
verify_hwid_checksum(char *tail_buf, char *hwid_fwinfo)
{
	char *hwid_bootblock;
	char hwid_image[HWID_LIST_LEN + 1];
	char *hwid;
	char *space;
	char *colon;
	char *dot;

	D(msg(__FUNCTION__));

	/*
	 * Fetch the hwid from the bootblock, the call may fail.
	 */
	hwid_bootblock = get_hwid_from_bootblock();
	if (hwid_bootblock == NULL) {
		msg(info_no_hwid);
	}
	D(if (hwid_bootblock) msg(hwid_bootblock));

	if (tail_buf) { /* csum check */
		char csum_string[CSUM_LEN + 1];
		unsigned long csum_image;
		unsigned long csum_calc;

		/* get image csum */
		strncpy(csum_string, tail_buf + HWID_LIST_LEN, CSUM_LEN);
		csum_string[CSUM_LEN] = '\0';
		csum_image = strtoul(csum_string, NULL, 10);

		/* get image hwid */
		strncpy(hwid_image, tail_buf, HWID_LIST_LEN);
		hwid_image[HWID_LIST_LEN] = '\0';

		/* compare checksums */
		csum_calc = calc_checksum(flash_layout, hwid_image);
		D(msg_int("csum_image", csum_image));
		if (csum_calc != csum_image) {
			msg(error_checksum_wrong);
			return -1;
		}
	} else { /* no tail_buf, get hwid from hwid_fwinfo */
		memcpy(hwid_image, hwid_fwinfo, HWID_LIST_LEN);
		hwid_image[HWID_LIST_LEN] = '\0';
	}

	if (!hwid_bootblock) {
		/* Skip test. */
		return 0;
	}

	/*
	 * If HWID from bootblock contains '.' it belongs to a range and we
	 * must only use the characters before the '.' for comparison.
	 */
	if ((dot = strchr(hwid_bootblock, '.'))) {
		*dot = '\0';
	}

	/* Truncate HWID from image at first space if any. */
	if ((space = strchr(hwid_image, ' '))) {
		*space = '\0';
	}

	/* The image may have several HWIDs separated by colon. */
	hwid = hwid_image;
	while ((colon = strchr(hwid, ':'))) {
		*colon = '\0';
		if (strncasecmp(hwid, hwid_bootblock, HWID_LIST_LEN) == 0) {
			return 0;
		}
		hwid = colon + 1;
	}
	if (strncasecmp(hwid, hwid_bootblock, HWID_LIST_LEN) == 0) {
		return 0;
	}

	/* Not found. */
	msg(error_hwid);
	return -1;
}

/* Free the dynamically allocated memory used for image file storage. */
void
free_iov_buffers(void)
{
	struct partition_layout *partition;
	int i;

	D(msg(__FUNCTION__));

	for (partition = flash_layout; partition->flash_len; partition++) {
		if (partition->iov != NULL) {
			for (i = 0; i < partition->iov_length; i++) {
				free(partition->iov[i].iov_base);
			}
			free(partition->iov);
			partition->iov = 0;
			partition->iov_length = 0;
		}
	}
}

/* Return 0 if it is possible to allocate the buffers defined in
   partition_layout, return -1 if malloc fails. The partition_layout
   contains an initialized buffer pointer area with allocated memory
   if 0 is returned. */
int
init_iov_buffers_p(struct partition_layout *partition)
{
	int no_iovs;
	int i;

	D(msg(__FUNCTION__));

	/* Reset iov buffers. */
	partition->iov = 0;
	partition->iov_length = 0;

	/*
	 * Allocate the iov buffers and store the length of each
	 * buffer in the buffer pointer area. Panic if malloc fails.
	 */
	/*
	 * Calculate the number of buffers required considering
	 * overhead which may leave a shorter buffer at the end.
	 */
	no_iovs = partition->image_len / partition->erasesize;
	D(msg_int("no_iovs", no_iovs));

	if (partition->image_len % partition->erasesize) {
		msg(error_modulo);
		return -1;
	}

	if (no_iovs == 0) {
		return 0;
	}

	/*
	 * Allocate the iov buffer pointer area or panic.
	 */
	partition->iov = malloc(no_iovs * sizeof(*(partition->iov)));
	if (partition->iov == NULL) {
		errmsg_errno(error_malloc, errno);
		return -1;
	}
	memset(partition->iov, 0, no_iovs * sizeof(*(partition->iov)));
	partition->iov_length = no_iovs;
	/*
	 * Allocate the buffers.
	 */
	for (i = 0; i < no_iovs; i++) {
		partition->iov[i].iov_base = malloc(partition->erasesize);
		if (partition->iov[i].iov_base == NULL) {
			errmsg_errno(error_malloc, errno);
			return -1;
		}
		partition->iov[i].iov_len = partition->erasesize;
	}

	return 0;
}

/* Return 0 if the rescue sector length can be read in flash memory or -1. */
static int
read_rescue_sector_length(unsigned int *length)
{
	int fd;
	mtd_info_t mtd_info;

	D(msg(__FUNCTION__));

	/* Open the rescue partition in flash memory. */
	if ((fd = open(dev_name_rescue, O_RDONLY)) < 0) {
		errmsg_errno(error_open, errno);
		goto panic;
	}

	if (get_mtd_info(fd, &mtd_info) < 0) {
		errmsg_errno(error_ioctl, errno);
		goto panic;
	}

	D(msg_int("rescue length", mtd_info.size));

	if (close(fd) < 0) {
		errmsg_errno(error_close, errno);
		goto panic;
	}
	*length = mtd_info.size;
	return 0;

panic:
	(void)close(fd);
	*length = 0;
	return -1;
}

/* Return 0 if a partition table is found in the supplied buffer,
   which should be a pointer to the beginning of the image kernel
   partition, or return -1. The size and type of each partition are
   stored in layout. */
int
read_image_partition_table(unsigned char *buf)
{
	struct pt_head *head;
	struct pt_entry *entry;
	int magic_start;
	int i;

	D(msg(__FUNCTION__));

	/* Search for magic */
	for (i = 0; i < MAX_MAGIC_OFFSET; i++) {
		if (*(((unsigned short*)buf) + i) == PTABLE_MAGIC)
			break;
	}
	if (i == MAX_MAGIC_OFFSET) {
		msg(error_find_ptable);
		return -1;
	}
	magic_start = i * sizeof(unsigned short);

	/* Read defined partition table entries from the buffered image. */
	head = (struct pt_head *)(buf + magic_start);
	i = KERNEL_PARTITION;
	entry = (struct pt_entry *)(buf + magic_start + sizeof(struct pt_head));
	do {
		flash_layout[i].image_len    = entry->size;
		flash_layout[i].image_offset = entry->offset;
		flash_layout[i].image_type   = entry->type;
		flash_layout[i].next         = &flash_layout[i+1];
		D(msg_int("partition_nr",    i));
		D(msg_int("entry->offset",   entry->offset));
		D(msg_int("entry->checksum", entry->checksum));
		D(msg_int("entry->type",     entry->type));
		D(msg_int("entry->flags",    entry->flags));
		D(msg_int("entry->size",     entry->size));
		++i;
		++entry;
	} while (entry->size > 0);

	return 0;
}

/* Return 0 if a partition table is found in the flash memory kernel partition
   or return -1. The size and type of each partition are stored in layout. */
int
read_flash_partition_table(void)
{
	unsigned int rescue_length;
	int fd = -1;
	ssize_t rbyte;
	int i;
	struct pt_head head;
	struct pt_entry entry;
	unsigned short data;
	mtd_info_t mtd_info;

	D(msg(__FUNCTION__));

	/* The initial rescue sector length is the offset to the partition
	 * table sector. It is not stored in the partition tables. */
	if (read_rescue_sector_length(&rescue_length) < 0) {
		goto panic;
	}

	flash_layout[RESCUE_PARTITION].flash_len = rescue_length;
	flash_layout[RESCUE_PARTITION].image_len = rescue_length;

	/* Open the kernel partition in flash memory. */
	if ((fd = open(dev_name_kernel, O_RDONLY)) < 0) {
		errmsg_errno(error_open, errno);
		goto panic;
	}

	if (get_mtd_info(fd, &mtd_info) < 0) {
		errmsg_errno(error_ioctl, errno);
		goto panic;
	}

#ifdef NANDFLASH_SUPPORT
	/* find next working block */
	if (mtd_info.type == MTD_NANDFLASH) {
		loff_t flash_addr = 0;
		int blockbad;

		do {
			if ((blockbad = ioctl(fd, MEMGETBADBLOCK, &flash_addr)) < 0) {
				errmsg_errno(error_ioctl, errno);
				msg_int("read_flash_partition_table ioctl mtd MEMGETBADBLOCK", flash_addr);
				goto panic;
			}
			if (blockbad) {
				/* block was bad, lseek to new position to synchronize */
				D(msg_int("Skipping bad block, offset", flash_addr));
				flash_addr += mtd_info.erasesize;
				if (lseek(fd, flash_addr, SEEK_SET) == -1) {
					errmsg_errno(error_lseek, errno);
					goto panic;
				}
			}
		} while (blockbad);
	}
#endif

	/* Search for magic */
	for (i = 0; i < MAX_MAGIC_OFFSET; i++) {
		read(fd, &data, sizeof(data));
		if (data == PTABLE_MAGIC) {
			msg("found magic\n");
			break;
		}
	}

	if (lseek(fd, -(off_t)sizeof(data), SEEK_CUR) == -1) {
		errmsg_errno(error_lseek, errno);
		goto panic;
	}

	/* Read the partition table head. */
	memset(&head, 0, sizeof(struct pt_head));
	rbyte = read(fd, &head, sizeof(struct pt_head));
	if (rbyte != sizeof(struct pt_head)) {
		errmsg_errno(error_read, errno);
		goto panic;
	}
	D(msg_int("head.magic", head.magic));
	D(msg_int("head.size", head.size));
	D(msg_int("head.checksum", head.checksum));

	/* Read defined partition table entries from flash
	   memory. There is no rescue partition. */
	i = KERNEL_PARTITION;
	do {
		memset(&entry, 0, sizeof(struct pt_entry));
		rbyte = read(fd, &entry, sizeof(struct pt_entry));
		if (rbyte != sizeof(struct pt_entry)) {
			goto panic;
		}
		if (entry.size > 0) {
			flash_layout[i].flash_len = entry.size;
			flash_layout[i].image_len = UNKNOWN_LEN;
			flash_layout[i].flash_type = entry.type;
			flash_layout[i].image_type = PARTITION_TYPE_UNDEF;
			D(msg_int("\npartition_nr", i));
			D(msg_int("entry.offset", entry.offset));
			D(msg_int("entry.checksum", entry.checksum));
			D(msg_int("entry.type", entry.type));
			D(msg_int("entry.flags", entry.flags));
			D(msg_int("entry.size", entry.size));
			++i;
		}
	} while (entry.size > 0);

	if (close(fd) < 0) {
		errmsg_errno(error_close, errno);
		goto panic;
	}

	return 0;

panic:
	if (fd != -1) {
		(void)close(fd);
	}
	return -1;
}


/* Return 0 if the image and flash partition tables are of equal lengths and
   types or return -1. */
int
compare_partition_tables(void)
{
	struct partition_layout *partition;

	D(msg(__FUNCTION__));

	for (partition = flash_layout; partition->flash_len; partition++) {
		D(msg_int("image_len", partition->image_len));
		D(msg_int("flash_len", partition->flash_len));
		D(msg_int("image_type", partition->image_type));
		D(msg_int("flash_type", partition->flash_type));
		if (partition->image_len != partition->flash_len
		    || partition->image_type != partition->flash_type) {
			msg(error_compare_ptable);
			return -1;
		}
	}
	return 0;
}

/* Print all percentages from last percentage if percent differs from
   last percent. Counter is reset and no percentage printed if called
   with 0 percent. I.e not reentrant safe. */
static void
print_percent(int percent, enum flash_protocol protocol)
{
	static int messages = 0;

	if (percent == 0) {
		messages = 1;
		return;
	}

	if (messages != percent) {
		while (messages <= percent && messages <= 100) {
			fprintf(stdout, " %3d%% ", messages);
			if (protocol == METHOD_HTTP_POST &&
			    (messages % 12 == 0)) {
				fprintf(stdout, "\r\n");
			} else if (protocol == METHOD_FTP) {
				fprintf(stdout, "\r\n");
			}
			messages++;
		}
		fflush(stdout);
	}
}

/* Return 0 if the partition can be erased, return -1 on failure. Erasure is
   attempted 10 times for a sector on nor flash. */
static int
erase_flash(struct partition_layout *partition, enum flash_protocol protocol)
{
	mtd_info_t mtd_info;
	erase_info_t mtd_erase;
	unsigned long size;

	D(msg(__FUNCTION__));

	if (partition->fd < 0) {
		msg("erase_flash:device not open");
		errmsg_errno(error_open, errno);
		goto panic;
	}
	if (lseek(partition->fd, 0, SEEK_SET) == -1) {
		errmsg_errno(error_lseek, errno);
		goto panic;
	}

	if (get_mtd_info(partition->fd, &mtd_info) < 0) {
		errmsg_errno(error_ioctl, errno);
		goto panic;
	}

	msg(info_erasing);

	D(msg("Erasing ..."));
	msg(partition->device_name);

	size = 0;

	print_percent(0, protocol);

	{
		unsigned blocks = mtd_info.size/ mtd_info.erasesize;
		unsigned b;
		int blockbad = 0;

		for (b = 0; b < blocks; b++) {
			int attempts;

			memset(&mtd_erase, 0, sizeof(erase_info_t));
			mtd_erase.start = b * mtd_info.erasesize;
			mtd_erase.length = mtd_info.erasesize;

#ifdef NANDFLASH_SUPPORT
			{
				loff_t flash_addr = mtd_erase.start;

				if (mtd_info.type == MTD_NANDFLASH &&
				    (blockbad = ioctl(partition->fd, MEMGETBADBLOCK, &flash_addr)) < 0) {
					errmsg_errno(error_ioctl, errno);
					msg_int("erase ioctl mtd MEMGETBADBLOCK", mtd_erase.start);
					goto panic;
				}
			}
#endif

			if (blockbad) {
				D(msg_int("Skipping bad block at offset", b*mtd_info.erasesize));
				continue;
			}

			attempts = 0;
			do {
				if ((ioctl(partition->fd, MEMERASE, &mtd_erase)) < 0) {
					errmsg_errno(error_ioctl, errno);
					msg(error_erase);
					if (++attempts > 10) {
						goto panic;
					}
				}
			} while (attempts && (mtd_info.type == MTD_NORFLASH));
			print_percent((b*100)/blocks, protocol);
		}
	}

	print_percent(100, protocol);

	fprintf(stdout, "\r\n");
	fflush(stdout);

	return 0;

panic:
	return -1;
}

/* Iterator for image iovectors. Return pointer to next or current iov
   from image, depending on cmd, or NULL if no next iov. If valid iov,
   fills in info in cur_iov struct. Starts in KERNEL_PARTITION. Not
   reentrant safe. */
static struct iovec *
next_iov(unsigned cmd, struct cur_iov *cur_iov)
{
//	static unsigned partition_nr = KERNEL_PARTITION;
//	static unsigned iov_nr       = 0;
	struct partition_layout *p   = &flash_layout[cur_iov->partition_nr];
	struct iovec *iov;

	if (cmd == CMD_NEXT_IOV) {
		cur_iov->iov_nr++;
	}

	/* next partition? but don't move past last partition */
	if ((cur_iov->iov_nr >= p->iov_length) && (p->image_len)) {
		D(msg("next_iov: next i partition"));
		cur_iov->partition_nr++;
		p = &flash_layout[cur_iov->partition_nr];
		D(msg_int("next_iov: blocks_in_i_partition", p->image_len/p->erasesize));
		D(msg_int("next_iov: iov", (unsigned)p->iov));
		D(msg_int("next_iov: partition_nr", cur_iov->partition_nr));
		cur_iov->iov_nr = 0;
	}

	/* end of image? */
	if (p->image_len == 0) {
		D(msg("next_iov: end of image\n"));
		return NULL;
	}

	/* fill in struct info */
	cur_iov->blocks_in_i_partition = p->image_len/p->erasesize;
	cur_iov->block_nr_in_i_partition = cur_iov->iov_nr;
	cur_iov->next = p->next;

	iov = &p->iov[cur_iov->iov_nr];

	return iov;
}

static int
skip1(struct cur_iov *cur_iov, struct iovec *iov,
      struct partition_layout *partition)
{
	while (cur_iov->skip > 0) {
		cur_iov->block_nr_in_f_partition++;
		cur_iov->flash_addr += iov->iov_len;
		cur_iov->part_addr += iov->iov_len;
		cur_iov->skip -= iov->iov_len;
		D(msg_int("skip1 to flash_addr", cur_iov->flash_addr));
		/* end of flash partition? */
		if ((cur_iov->block_nr_in_f_partition == cur_iov->blocks_in_f_partition)) {
			D(msg_int("end of flash partition in i skip", cur_iov->skip));
			return 0;
		} else {
			D(msg("lseek"));
			if (lseek(partition->fd, iov->iov_len, SEEK_CUR) == -1) {
				errmsg_errno(error_lseek, errno);
				return -1;
			}
		}
	}

	return 1;
}

/* Return 0 if the data from the image stored in partition can be
   written to flash partition, return -1 if write fails. Each call to
   this function programs the entire flash partition. Not reentrant
   safe. */
static int
program_flash(struct partition_layout *partition, enum flash_protocol protocol)
{
	int tot_iovs;
	mtd_info_t mtd_info;

	D(msg(__FUNCTION__));
	D(msg("Programming ..."));

	if (partition->fd < 0) {
		msg("program_flash:device not open");
		errmsg_errno(error_open, errno);
		goto panic;
	}
	if (lseek(partition->fd, 0, SEEK_SET) == -1) {
		errmsg_errno(error_lseek, errno);
		goto panic;
	}

	msg(info_loading);
	msg(partition->device_name);

	tot_iovs = 0;

	if (get_mtd_info(partition->fd, &mtd_info) < 0) {
		errmsg_errno(error_ioctl, errno);
		goto panic;
	}

	{
		int blockbad = 0;
		static struct cur_iov cur_iov = { 0, };
		struct iovec *iov = NULL;
		unsigned erasesize;
		unsigned int len;
		int ret;

		if (cur_iov.flash_addr == 0) {
			cur_iov.partition_nr = KERNEL_PARTITION;
			cur_iov.flash_addr = flash_layout[RESCUE_PARTITION].flash_len;
		}


		/* initalizations */
		len                             = mtd_info.size;
		erasesize                       = mtd_info.erasesize;
		cur_iov.blocks_in_f_partition   = mtd_info.size/erasesize;
		cur_iov.block_nr_in_f_partition = 0;
		cur_iov.part_addr               = 0;

		/* get current iov */
		iov = next_iov(CMD_CURRENT_IOV, &cur_iov);

		D(msg_int("blocks_in_i_partition", cur_iov.blocks_in_i_partition));
		D(msg_int("blocks_in_f_partition", cur_iov.blocks_in_f_partition));

		print_percent(0, protocol);

		/* loop through flash partition */
		while ((cur_iov.block_nr_in_f_partition < cur_iov.blocks_in_f_partition) &&
		       iov) {
			unsigned long bytes_to_write;

			D(msg_int("block_nr_in_i_partition", cur_iov.block_nr_in_i_partition));
			D(msg_int("block_nr_in_f_partition", cur_iov.block_nr_in_f_partition));
			D(msg_int("tot_iovs", tot_iovs));
			D(msg_int("flash_addr", cur_iov.flash_addr));
			D(msg_int("part_addr", cur_iov.part_addr));

			/* Too small partition? This could happen if we get too
			   many bad blocks in a NAND */
			if (cur_iov.part_addr > mtd_info.size) {
				msg(error_eop);
				goto panic;
			}

			bytes_to_write = iov->iov_len;
			D(msg_int("bytes_to_write", bytes_to_write));

			/* skip space between partitions? */
			ret = skip1(&cur_iov, iov, partition);
			if (ret == -1) {
				goto panic;
			} else if (ret == 0) {
				return ret;
			}

#ifdef NANDFLASH_SUPPORT
			/* Check for (already) bad block */
			if (mtd_info.type == MTD_NANDFLASH) {
				blockbad = ioctl(partition->fd, MEMGETBADBLOCK, &cur_iov.part_addr);
				if (blockbad < 0) {
					msg_int("program_flash(1) ioctl mtd MEMGETBADBLOCK", cur_iov.part_addr);
					goto panic;
				}
			}
#endif

			/* Write if block is ok before we start */
			if (!blockbad) {
				int wrstatus;

				D(msg_int("write", cur_iov.flash_addr));
				wrstatus = write(partition->fd, iov->iov_base, iov->iov_len);

				if (wrstatus != iov->iov_len) {
					/* write failed */
					if (errno == EIO) {
#ifdef NANDFLASH_SUPPORT
						if (mtd_info.type == MTD_NANDFLASH) {
							/* Block became bad during write. Mark it. */
							if (ioctl(partition->fd, MEMSETBADBLOCK, &cur_iov.part_addr) < 0) {
								msg("program_flash(2) ioctl mtd MEMSETBADBLOCK");
								goto panic;
							}
							blockbad = 1;
							msg_int("Block failed during write, offset", cur_iov.part_addr);
						}
#endif
					} else {
						errmsg_errno(error_writev, errno);
						goto panic;
					}
				} else {
					/* write() successful, check length of write to be sure */
					if (wrstatus != bytes_to_write) {
						errmsg_errno(error_writev, errno);
						D(msg_int("Write failure at partiton address\n", cur_iov.part_addr));
						D(msg_int("wrstatus", wrstatus));
						D(msg_int("bytes_to_write", bytes_to_write));
						goto panic;
					}
					len -= bytes_to_write;
				}
			}

			cur_iov.block_nr_in_f_partition++;
			cur_iov.part_addr  += bytes_to_write;
			cur_iov.flash_addr += bytes_to_write;

			if (blockbad) {
				/* block was bad or became bad, lseek to new position */
				if (lseek(partition->fd, iov->iov_len, SEEK_CUR) == -1) {
					errmsg_errno(error_lseek, errno);
					goto panic;
				}
				D(msg("lseek"));
				D(msg_int("Skipping bad block, offset", cur_iov.part_addr));
			} else { /* block write ok */
				tot_iovs++;

				/* move if more in this flash partition */
				/* end of image partition? skip free space */
				if (cur_iov.block_nr_in_i_partition == cur_iov.blocks_in_i_partition-1) {
					D(msg("\nend of image partition"));

					/* skip? */
					if (cur_iov.next && cur_iov.next->image_offset) {
						unsigned next_offset = cur_iov.next->image_offset +
							flash_layout[RESCUE_PARTITION].flash_len;
						D(msg_int("next offset", next_offset));
						D(msg_int("flash_addr", cur_iov.flash_addr));

						cur_iov.skip = next_offset - cur_iov.flash_addr;
						D(msg_int("skip", cur_iov.skip));
					}

					/* do the skip */
					ret = skip1(&cur_iov, iov, partition);
					if (ret == -1) {
						goto panic;
					} else if (ret == 0) {
						iov = next_iov(CMD_NEXT_IOV, &cur_iov);
						print_percent(100, protocol);
						return ret;
					}

					cur_iov.skip = 0;
				}

				iov = next_iov(CMD_NEXT_IOV, &cur_iov);
			}

			print_percent(((tot_iovs * 100) / partition->iov_length), protocol);
		} /* while */

		print_percent(100, protocol);

		if (cur_iov.block_nr_in_f_partition == cur_iov.blocks_in_f_partition) {
			D(msg("end of flash partition"));
		}
		if (!iov) {
			D(msg("end of image"));
		}

		fprintf(stdout, "\r\n");
		fflush(stdout);
	}

	return 0;

panic:
	return -1;
}

/* Return 0 if the content of flash memory equals that of the image written to
   flash memory, return -1 on failure or on differences. */
static int
verify_flash(struct partition_layout *partition)
{
	unsigned char *buf = NULL;
	int byte_r;
	mtd_info_t mtd_info;
	struct iovec *iov;
	static struct cur_iov cur_iov = { 0, };
	int ret;

	if (cur_iov.flash_addr == 0) {
		cur_iov.partition_nr = KERNEL_PARTITION;
		cur_iov.flash_addr = flash_layout[RESCUE_PARTITION].flash_len;
	}

	if ((iov = next_iov(CMD_CURRENT_IOV, &cur_iov)) == NULL) {
		return 0;
	}

	D(msg(__FUNCTION__));
	D(msg("Verifying ..."));
	D(msg(partition->device_name));

	if ((buf = malloc(partition->erasesize)) == NULL) {
		errmsg_errno(error_open, errno);
		goto panic;
	}

	if (partition->fd < 0) {
		msg("verify_flash:device not open");
		errmsg_errno(error_open, errno);
		goto panic;
	}

	if (lseek(partition->fd, 0, SEEK_SET) == -1) {
		msg("verify_flash: lseek to file start");
		errmsg_errno(error_lseek, errno);
		goto panic;
	}

	if (get_mtd_info(partition->fd, &mtd_info) < 0) {
		errmsg_errno(error_ioctl, errno);
		goto panic;
	}

	/* initalizations */
	cur_iov.blocks_in_f_partition   = mtd_info.size/mtd_info.erasesize;
	cur_iov.block_nr_in_f_partition = 0;
	cur_iov.part_addr               = 0;

	/* skip space between partitions? */
	ret = skip1(&cur_iov, iov, partition);
	if (ret == -1) {
		goto panic;
	} else if (ret == 0) {
		return ret;
	}

	while (cur_iov.skip > 0) {
		cur_iov.block_nr_in_f_partition++;
		cur_iov.flash_addr+=iov->iov_len;
		cur_iov.part_addr+=iov->iov_len;
		cur_iov.skip-=iov->iov_len;
		D(msg_int("skip1 to flash_addr", cur_iov.flash_addr));
		/* end of flash partition? */
		if ((cur_iov.block_nr_in_f_partition == cur_iov.blocks_in_f_partition)) {
			D(msg_int("end of flash partition in i skip", cur_iov.skip));
			return 0;
		} else {
			if (lseek(partition->fd, iov->iov_len, SEEK_CUR) == -1) {
				errmsg_errno(error_lseek, errno);
				goto panic;
			}
			D(msg("lseek"));
		}
	}

	/* loop over flash partition, but not longer than image partition */
	while (cur_iov.block_nr_in_f_partition < cur_iov.blocks_in_f_partition  && iov) {
		int blockbad = 0;

		D(msg_int("block_nr_in_i_partition", cur_iov.block_nr_in_i_partition));
		D(msg_int("block_nr_in_f_partition", cur_iov.block_nr_in_f_partition));

		if (cur_iov.part_addr > mtd_info.size) {
			msg("Out of memory in partition");
			goto panic;
		}

#ifdef NANDFLASH_SUPPORT
		if (mtd_info.type == MTD_NANDFLASH) {
			blockbad = ioctl(partition->fd, MEMGETBADBLOCK, &cur_iov.part_addr);
			if (blockbad < 0) {
				errmsg_errno(error_ioctl, errno);
				msg("verify_flash ioctl mtd MEMGETBADBLOCK");
				goto panic;
			}
		}
#endif

		if (blockbad) {
			D(msg_int("Skipping bad block at part_addret", cur_iov.part_addr));
			if (lseek(partition->fd, iov->iov_len, SEEK_CUR) == -1) {
				errmsg_errno("error in lseek", errno);
				goto panic;
			}
		} else { /* block not bad */
			byte_r = read(partition->fd, buf, iov->iov_len);
			D(msg_int("cur_iov.part_addr", cur_iov.part_addr));
			D(msg_int("flash_addr", cur_iov.flash_addr));
			if (byte_r < 0 || byte_r != iov->iov_len) {
				errmsg_errno(error_read, errno);
				goto panic;
			}
			if (memcmp(buf, iov->iov_base, iov->iov_len) != 0) {
				msg(error_verify);
				goto panic;
			}
		}

		/* move */
		if (iov) { /* iov can be null here */
			cur_iov.part_addr += iov->iov_len;
			cur_iov.flash_addr += iov->iov_len;
		}
		cur_iov.block_nr_in_f_partition++;

		if (!blockbad) { /* only if we moved in flash */
			if (cur_iov.block_nr_in_i_partition == cur_iov.blocks_in_i_partition-1) {
				D(msg("\nend of image partition"));

				/* skip? */
				if (cur_iov.next && cur_iov.next->image_offset) {
					unsigned next_offset =
						cur_iov.next->image_offset +
						flash_layout[RESCUE_PARTITION].flash_len;
					D(msg_int("next offset", next_offset));
					D(msg_int("part_addr", cur_iov.part_addr));

					cur_iov.skip = next_offset - cur_iov.flash_addr;
					D(msg_int("skip", cur_iov.skip));
				}

				/* do the skip */
				ret = skip1(&cur_iov, iov, partition);
				if (ret == -1) {
					goto panic;
				} else if (ret == 0) {
					iov = next_iov(CMD_NEXT_IOV, &cur_iov);
					return ret;
				}

				cur_iov.skip = 0;
			} /* end of image partition */

			iov = next_iov(CMD_NEXT_IOV, &cur_iov);
		} /* block not bad */
	}

	free(buf);
	return 0;

panic:
	free(buf);
	return -1;
}

/* Return 0 if the device for the flash memory partition was opened (O_RDWR)
   or return -1. */
int
open_flash_device(struct partition_layout *partition)
{
	D(msg(__FUNCTION__));

	/* Open the partition read/write in flash memory. */
	if (partition->fd == -1) {
		if ((partition->fd = open(partition->device_name, O_RDWR)) < 0) {
			partition->fd = -1;
			errmsg_errno(error_open, errno);
			msg(partition->device_name);
			return -1;
		}
	}

	return 0;
}

/* Close open file descriptors. */
void
close_flash_devices(struct partition_layout *partition[],
                    enum flash_protocol protocol)
{
	int ix;

	D(msg(__FUNCTION__));

	for (ix = 0; ix < nr_of_partitions; ix++) {
		if (partition[ix] && (partition[ix]->fd >= 0)) {
			close(partition[ix]->fd);
			partition[ix]->fd = -1;
		}
	}
}

/* Get mtd_info. Return -1 on fail. */
static int
get_mtd_info(int fd, mtd_info_t *mtd_info)
{
	memset(mtd_info, 0, sizeof(mtd_info_t));

	if (ioctl(fd, MEMGETINFO, mtd_info) < 0) {
		return -1;
	}

	return 0;
}

/* Return 0 if a flash memory partition can be erased, programmed and verified
   against the image data programed by or return -1. */
int
update_flash(struct partition_layout *partition, enum flash_protocol protocol)
{
	D(msg(__FUNCTION__));
	D(msg(partition->device_name));

	if (erase_flash(partition, protocol) < 0) {
		return -1;
	}
	D(msg("erase_flash: OK"));

	if (program_flash(partition, protocol) < 0) {
		return -1;
	}
	D(msg("program_flash: OK"));

	{
		mtd_info_t mtd_info;

		/* Obtain information on the partition in flash memory. */
		if (get_mtd_info(partition->fd, &mtd_info) < 0) {
			errmsg_errno(error_ioctl, errno);
			exit(EXIT_FAILURE);
		}
//		if (mtd_info.type == MTD_NORFLASH) {
			if (verify_flash(partition) < 0) {
				return -1;
			}
			D(msg("verify_flash: OK"));
//		}
	}

	return 0;
}
