/*
 * flash.h
 *
 * (C) Copyright 2002-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef __FLASH_H__
#define __FLASH_H__

#include <sys/uio.h>

enum partition {
	RESCUE_PARTITION,
	KERNEL_PARTITION,
};

/* HTTP and FTP require different output formats. */
enum flash_protocol {
	METHOD_HTTP_POST,
	METHOD_FTP,
	METHOD_CMDLINE,
	NO_PROTOCOL
};

enum tail_format {
	HWID_LIST_LEN = 8,
	CSUM_LEN = 16,
	TAIL_LEN = HWID_LIST_LEN + CSUM_LEN
};

struct partition_layout {
	unsigned int flash_len;	/* Length of partition in flash memory. */
	unsigned int image_len;	/* Length of partition in image. */
	unsigned int image_offset; /* Offset from start of flash. */
	const char *device_name;/* Device name. */
	unsigned short flash_type; /* Partition type in flash. */
	unsigned short image_type; /* Partition type in image. */
	struct iovec *iov;	/* Vector of buffers. */
	int iov_length;		/* Number of buffers in iov. */
	int fd;			/* The O_RDWR opened file descriptor for
				 * the device, -1 if not open. */
	int erasesize;
	struct partition_layout *next;
};

struct partition_layout *flash_layout;

extern int nr_of_partitions;

/* Return 0 if a partition table is found in the flash memory kernel partition
   or return -1. The size and type of each partition are stored in layout. */
extern int read_flash_partition_table(void);

/* Return 0 if it is possible to allocate the buffers defined in flash_layout,
   return -1 if malloc fails. The flash_layout contains an initialized
   buffer pointer area with allocated memory if 0 is returned. */
extern int init_iov_buffers(struct partition_layout layout[]);

/* Return 0 if it is possible to allocate the buffers defined in
   partition_layout, return -1 if malloc fails. The partition_layout
   contains an initialized buffer pointer area with allocated memory
   if 0 is returned. */
extern int init_iov_buffers_p(struct partition_layout *layout);

/* Free the dynamically allocated memory used for image file storage. */
extern void free_iov_buffers(void);

/* Return 0 if the HWID and checksum of the image are valid, return -1 if
   either or both are invalid. */
extern int verify_hwid_checksum(char *tail_buf, char *fwinfo_hwid);

/* Return 0 if a partition table is found in the supplied buffer,
   which should be a pointer to the beginning of the image kernel
   partition, or return -1. The size and type of each partition are
   stored in layout. */
extern int read_image_partition_table(unsigned char *buf);

/* Return 0 if the image and flash partition tables are of equal lengths and
   types or return -1. */
extern int compare_partition_tables(void);

/* Return 0 if the device for the flash memory partition was opened (O_RDWR)
   or return -1. */
extern int open_flash_device(struct partition_layout *partition);

/* Close open file descriptors. */
extern void close_flash_devices(struct partition_layout *partition[],
			 enum flash_protocol protocol);

extern int update_flash(struct partition_layout *partition,
			enum flash_protocol protocol);

/* Return 0 if paging is disabled for all pages mapped into the address space
   of the calling process, return -1 on failure. */
extern int disable_paging(void);

/* Return blocksize/eraseize of partition, or -1 on fail */
extern int get_block_size(struct partition_layout *partition);

/* return nr of partitions in flash, -1 on fail */
extern int get_nr_of_partitions();

/* Set length of an image partition in the flash_layout. */
extern void set_length(unsigned partition_nr, unsigned long len);

/* Return length of a flash partition from the flash_layout. */
unsigned long get_length(unsigned partition_nr);

#endif /* __FLASH_H__ */
