/*
 * main.c
 *
 * This application is used to perform NOR and NAND flash memory
 * upgrades with linux running.
 *
 * (C) Copyright 2002-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/reboot.h>
#include <linux/reboot.h>
#include <signal.h>
#include <fwinfo.h>
#include "flash.h"
#include "msg.h"
#include "external_call.h"
#include "read_image.h"

/* define for debug printouts */
#define D(x)

/* Mapped to flash targets. */
enum flash_target {
	TARGET_FLASH     = 0,
	TARGET_FLASH_ALL = 1,
	NO_TARGET        = 2
};

/* Terminate by exit(3) until the preparation script executes. */
enum error_condition {
	TERMINATE_BY_EXIT,
	TERMINATE_BY_REBOOT
};

/* Lock file name. */
const char lock_file[] = "/tmp/_flash__lock";

/* Command line options. */
static char opt_m_http_post[] = "http_post";
static char opt_m_ftp[] = "ftp";
static char opt_m_cmdline[] = "cmdline";
static char opt_t_flash[] = "flash";
static char opt_t_flash_all[] = "flash_all";

/* Form data terminators in HTTP POST. */
const char form_data_header_end[] = "\r\n\r\n";
const char form_data_trailer_end[] = "--\r\n";

/* Messages sent to the client. */
static const char error_sizes[] = "Error: cannot get partition sizes.";
static const char error_open[] = "Error on open.";
static const char error_close[] = "Error on close.";
static const char error_format[] = "Error: format error in data received.";
static const char error_script[] = "Error: script failed.";
static const char error_fwinfo[] = "Error: no fwinfo in image.";
static const char info_prescript[] = "Preparing system for upgrade ...";
static const char info_reboot_now[] =
	"The unit will now reboot.\r\n"
	"To continue, please connect to the unit again.\r\n"
	"There may be a short delay before the new connection is accepted.\r\n"
	"This connection will now close.";
static const char info_upgrading[] = "The system is already being upgraded.";
static const char info_success[] =
	"The system upgrade completed successfully.";
static const char info_failure[] = "The system upgrade failed.";

int nr_of_partitions;

/* Print usage. */
static void
usage(void)
{
	msg("syntax: flash [-h] -t <target> -m <method> [-f <file>]\r\n"
	    "   -h help.\r\n"
	    "   -t flash, flash_all.\r\n"
	    "   -m HTTP_POST, FTP.\r\n"
	    "   -f fimage (default; read from stdin).");
}

int
main(int argc, char **argv)
{
	enum error_condition terminate_condition;
	char *summary_message;
	enum flash_target target;
	char *target_string;
	enum flash_protocol protocol;
	char *protocol_string;
	int fd_lock;
	int datafd;
	int statfd = STDOUT_FILENO;
	char *filename;
	const char opt[] = "f:hm:s:t:";
	char buffer[TAIL_LEN];
	unsigned long old_rescue_len;
	int partitions_to_open;
	unsigned i;

	D(msg(__FUNCTION__));

	/* Any panic condition will result in a summary message of failure. */
	terminate_condition = TERMINATE_BY_EXIT;
	summary_message = (char *)info_failure;

	/* The target is an inparameter. */
	target = NO_TARGET;

	/* The target string based on the target is used in the script. */
	target_string= NULL;

	/* The method or protocol is an inparameter. */
	protocol = NO_PROTOCOL;

	/* The method string based on the method is used in the script. */
	protocol_string = NULL;

	/* The image file to read data from. */
	filename = NULL;
	datafd = -1;

	/* Check command line options. */
	if (argc < 2) {
		usage();
		goto panic;
	}

	for (;;) {
		int c = getopt(argc, argv, opt);

		if (c < 0)
			break;

		switch (c) {
		case 'f':
			filename = optarg;
			if ((datafd = open(filename, O_RDONLY)) < 0) {
				errmsg_errno(error_open, errno);
				goto panic;
			}
			break;
		case 'h':
			usage();
			goto panic;
			break;
		case 'm':
			if (strncasecmp(optarg, opt_m_http_post,
					strlen(optarg)) == 0) {
				protocol = METHOD_HTTP_POST;
				protocol_string = opt_m_http_post;
			} else if (strncasecmp(optarg, opt_m_ftp,
					       strlen(optarg)) == 0) {
				protocol = METHOD_FTP;
				protocol_string = opt_m_ftp;
				statfd = 3; /* by convention */
			} else if (strncasecmp(optarg, opt_m_cmdline,
					       strlen(optarg)) == 0) {
				protocol = METHOD_CMDLINE;
				protocol_string = opt_m_ftp;
			} else {
				usage();
				goto panic;
			}
			break;
		case 't':
			if (strncasecmp(optarg, opt_t_flash,
					strlen(optarg)) == 0) {
				target = TARGET_FLASH;
				target_string = opt_t_flash;
			} else if (strncasecmp(optarg, opt_t_flash_all,
					       strlen(optarg)) == 0) {
				target = TARGET_FLASH_ALL;
				target_string = opt_t_flash_all;
			} else {
				usage();
				goto panic;
			}
			break;
		default:
			break;
		}
	}

	D(fdmsg(statfd, "Parsed options."));

	/*
	 * Default to read data from stdin if no filename is defined.
	 */
	if (datafd < 0) {
		datafd = fileno(stdin);
	}

	/*
	 * Make sure a flash target is defined.
	 */
	if (target == NO_TARGET || protocol == NO_PROTOCOL) {
		usage();
		goto panic;
	}

	/* The upgrade messages must be displayed when sent. */
	if (protocol == METHOD_HTTP_POST) {
		msg("Content-Type: text/plain\r\n");
	}

	/* Lock out all other threads. */
	if ((fd_lock = open(lock_file, O_CREAT|O_EXCL|O_WRONLY, 0600)) == -1) {
		if (errno == EEXIST || errno == ENOENT) {
			fdmsg(statfd, info_upgrading);
		} else {
			errmsg_errno(error_open, errno);
			terminate_condition = TERMINATE_BY_REBOOT;
		}
		goto panic;
	}
	(void)close(fd_lock);
	/* Reboot on errors from now on. */
	terminate_condition = TERMINATE_BY_REBOOT;

	/*
	 * Execute a script which enters run level 4 and stores the
	 * type of upgrade requested. Give up on error.
	 */
	fdmsg(statfd, info_prescript);
	if (run_preparation_script(target_string, protocol_string,
				   statfd) < 0) {
		msg(error_script);
		goto panic;
	}

	if (statfd != STDOUT_FILENO) {
		close(statfd);
	}

	/* figure out how many partitions we have, and initialize stuff */
	if ((nr_of_partitions = get_nr_of_partitions()) < 0) {
		msg(error_open);
		goto panic;
	}

	/*
	 * Read the partition table stored in flash memory.
	 */
	if (read_flash_partition_table() < 0) {
		goto panic;
	}

	/*
	 * Lock to avoid page faults which are undesirable while erasing.
	 */
	if (disable_paging() < 0) {
		goto panic_free;
	}

	/*
	 * HTTP POST will send additional row oriented data in front
	 * of the image.
	 */
	if (protocol == METHOD_HTTP_POST) {
		if (read_additional_data(datafd, form_data_header_end) < 0) {
			msg(error_format);
			goto panic_free;
		}
	}

	/* Remember current rescue partition length, we have to change
	   it temporarily for NAND images. */
	old_rescue_len = get_length(RESCUE_PARTITION);

	/* initialize erasesize */
	for (i = RESCUE_PARTITION; i<nr_of_partitions; i++) {
		if (get_block_size(&flash_layout[i]) == -1) {
			msg(error_sizes);
			goto panic_free;
		}
	}

	{ /* read rescue partition, including fwinfo */
		unsigned char *fwinfo_buf;
		struct fwinfo fwinfo;

		if (!(fwinfo_buf = malloc(flash_layout[RESCUE_PARTITION].erasesize))) {
			errmsg_errno(error_close, errno);
			goto panic_free;
		}
		/* since we read some data from the rescue partition here
		   to try and find fwinfo we have to reuse this data and
		   copy it to the proper place in the image buffer
		   later */
		if (read_to_buffer(datafd, fwinfo_buf, flash_layout[RESCUE_PARTITION].erasesize) == -1) {
			goto panic_free;
		}

                /* get fwinfo from buf */
                if(read_fimage_fwinfo(fwinfo_buf, &fwinfo) == -1) {
                  msg(error_fwinfo);
                  goto panic_free;
                }

		verify_hwid_checksum(NULL, fwinfo.hwids);

		if (fwinfo.media_type == FWINFO_MEDIA_TYPE_NANDFLASH) {
			/* set the correct length of the rescue partition from
			   the length supplied in fwinfo */
			set_length(RESCUE_PARTITION, fwinfo.ptable_img_addr);
		}

		if (init_iov_buffers_p(&flash_layout[RESCUE_PARTITION]) == -1) {
			goto panic_free;
		}

		/* copy back the pre-read to rescue */
		memcpy(flash_layout[RESCUE_PARTITION].iov[0].iov_base, fwinfo_buf,
		       flash_layout[RESCUE_PARTITION].iov[0].iov_len);

		/* read rescue partition */
		if (read_image(datafd, &flash_layout[RESCUE_PARTITION], 1) == -1) {
			goto panic_free;
		}

		{ /* read kernel partition, including ptable */
			unsigned char *ptable_buf;

			if (!(ptable_buf = malloc(flash_layout[KERNEL_PARTITION].erasesize))) {
				errmsg_errno(error_close, errno);
				goto panic_free;
			}

			/* again a pre-read that we need to copy back below */
			if (read_to_buffer(datafd, ptable_buf, flash_layout[KERNEL_PARTITION].erasesize) == -1) {
				goto panic_free;
			}

			/* parse the ptable */
			if (read_image_partition_table(ptable_buf) == -1) {
				goto panic_free;
			}

			if (init_iov_buffers_p(&flash_layout[KERNEL_PARTITION]) == -1) {
				goto panic_free;
			}

			/* copy pre-read to kernel partition */
			memcpy(flash_layout[KERNEL_PARTITION].iov[0].iov_base, ptable_buf,
			       flash_layout[KERNEL_PARTITION].erasesize);

			/* read kernel partition */
			if (read_image(datafd, &flash_layout[KERNEL_PARTITION], 1) == -1) {
				goto panic_free;
			}
		}

		/* read the other partitions */
		for (i = KERNEL_PARTITION+1; i<nr_of_partitions; i++) {
			if (init_iov_buffers_p(&flash_layout[i]) == -1) {
				goto panic_free;
			}
			if (read_image(datafd, &flash_layout[i], 0) == -1) {
				goto panic_free;
			}
		}
	}

	/*
	 * Read the appended HWID and check sum of the image.
	 *
	 * The image contains at least 24 unsigned chars after the
	 * MB-alignment. The hardware id is 8 unsigned chars long and
	 * is padded with space from the right. The check sum is
	 * 16 unsigned chars long and is padded with space from the
	 * left, "%16ul". For example, 30A_____-------737204560
	 *
	 */

	if (read_to_buffer(datafd, buffer, TAIL_LEN) < 0) {
		msg(error_format);
		goto panic_free;
	}

	/*
	 * HTTP POST will send additional row oriented data after the image.
	 */

	if (protocol == METHOD_HTTP_POST) {
		if (read_additional_data(datafd, form_data_trailer_end) < 0) {
			msg(error_format);
			goto panic_free;
		}
	}

	/*
	 * Compare the HWID of the image with the one stored in flash memory
	 * and verify the check sum of the image.
	 */
	if (verify_hwid_checksum(buffer, NULL) < 0) {
		goto panic_free;
	}

	/* restore the temporarily changed RESCUE_PARTITION length
	   before flashing */
	set_length(RESCUE_PARTITION, old_rescue_len);

	/*
	 * Search for the partition table in the image sent.
	 */
	if (read_image_partition_table(flash_layout[KERNEL_PARTITION].iov[0].iov_base) < 0) {
		goto panic_free;
	}

	/*
	 * Compare the partition table of the image sent with that in
	 * flash memory for target "flash" only.
	 */
	if ((target == TARGET_FLASH) && (compare_partition_tables() < 0)) {
		goto panic_free;
	}

	/* Figure out how many partions to flash. */
	if (target == TARGET_FLASH_ALL) {
		partitions_to_open = nr_of_partitions-1; /* not rescue */
	} else {
		partitions_to_open = nr_of_partitions-2; /* not rescue and not last */
	}

	/*
	 * Open flash devices before erasing anything in the flash.
	 */
	for (i = KERNEL_PARTITION; i != (partitions_to_open+KERNEL_PARTITION); i++) {
		if (open_flash_device(&flash_layout[i]) < 0) {
			goto panic_free;
		}
	}

	/*
	 * Erase and program flash memory. Critical section for power failures.
	 */
	for (i = KERNEL_PARTITION; i != (partitions_to_open+KERNEL_PARTITION); i++) {
		if (update_flash(&flash_layout[i], protocol) < 0) {
			goto panic_free;
		}
	}

	/*
	 * Inform the user of successful completion.
	 */
	summary_message = (char *)info_success;

panic_free:
	D(msg("panic_free\n"));
	/*
	 * Free potential malloced memory.
	 */
	free_iov_buffers();

panic:
	/*
	 * Read all data from sending process.
	 */
	read_rest(datafd);

	msg(summary_message);
	if (terminate_condition == TERMINATE_BY_REBOOT) {
		msg(info_reboot_now);
	}

	if (close(datafd) < 0) {
		D(errmsg_errno(error_close, errno));
	}

	if (close(fileno(stdout)) < 0) {
		D(errmsg_errno(error_close, errno));
	}

	if (terminate_condition == TERMINATE_BY_REBOOT) {
		/*
		 * We need to careful here. Some processes may not get enough
		 * time to flush their message buffers if they don't handle
		 * the SIGTERM signal.
		 */

		/*
		 * Reboot! Either on successful upgrade or on fatal error.
		 */
		kill(-1, SIGTERM);
		sleep(5);

		kill(-1, SIGKILL);
		sleep(1);

		/* Ok, now pull the plug... */
		reboot(LINUX_REBOOT_CMD_RESTART);

	} else {
		/*
		 * Secondary threads exit due to upgrade in progress,
		 * or on errors which occur before the preparation script
		 * executes.
		 */
		exit(EXIT_FAILURE);
	}

	/* Never reached */
	exit(EXIT_SUCCESS);
}
