/*
 * read_image.c
 *
 * This module supports reading from a file descriptor in upgrades.
 *
 * (C) Copyright 2002-2007, Axis Communications AB, LUND, SWEDEN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <fwinfo.h>
#include "read_image.h"
#include "flash.h"
#include "msg.h"

/* define for debug printout */
#define D(x)

/* Available data after read, no data read but continue, and end of stream. */
enum read_code {
	READ_AVAILABLE_DATA,
	READ_NO_DATA,
	END_OF_DATA
};

/* Info messages. */
static const char info_receiving[] = "Receiving new firmware ...";

/* Error messages. */
static const char error_read[] = "Error: read failed.";
static const char error_fimage[] = "Error: too short file.";
static const char error_malloc[] = "Error: out of memory.";

/* Return readable status codes based on number of byte read by read. */
static enum read_code read_status(ssize_t byte_read);

/* Read data from datafd with timeout */
static int read_with_timeout(int fd, void *buf, size_t count, ssize_t *byte_r);

/* Read one partition from datafd into flash_layout. Return 0 on
   success, -1 on fail. */
int read_image(int datafd, struct partition_layout *partition, unsigned iov_offset);

/* Return 0 if buffer can be filled to len by reading from datafd or -1. */
int read_to_buffer(int datafd, char *buffer, int len);

/* Return 0 if an end_token is found within 2 kByte by reading one byte
   sequentially or return -1. */
int read_additional_data(int datafd, const char *end_token);

/* Read remaining data from datafd and ignore it. */
void read_rest(int datafd);

/* Return readable status codes based on number of byte read by read. */
static enum read_code
read_status(ssize_t byte_read)
{
	if (byte_read < 0) {
		if (errno == EINTR || errno == EAGAIN) {
			return READ_NO_DATA;
		} else {
			D(msg(__FUNCTION__));
			errmsg_errno(error_read, errno);
			return END_OF_DATA;
		}
	} else if (byte_read == 0) {
		D(msg(__FUNCTION__));
		msg(error_fimage);
		return END_OF_DATA;
	} else {
		return READ_AVAILABLE_DATA;
	}
}

/* Read one partition from datafd into flash_layout. Return 0 on
   success, -1 on fail. */
int
read_image(int datafd, struct partition_layout *partition, unsigned iov_offset)
{
	int i;
	unsigned char *buf;
	ssize_t byte_r;
	size_t leftr;
	enum read_code status;
	int tot;
	int sumbyte;		/* bytes read */

	D(msg(__FUNCTION__));

	msg(info_receiving);

	sumbyte = 0;
	for (i = iov_offset; i < partition->iov_length; i++) {
		buf = partition->iov[i].iov_base;
		leftr = partition->iov[i].iov_len;
		tot = 0;
		do {
			if (read_with_timeout(datafd, buf + tot, leftr, &byte_r) < 0) {
				msg(error_fimage);
				return -1; /* timed out */
			}
			status = read_status(byte_r);
			if (status == READ_NO_DATA) {
				continue;
			} else if (status == END_OF_DATA) {
				D(msg_int("failed sumbyte", sumbyte));
				msg(error_fimage);
				return -1;
			}
			leftr -= byte_r;
			tot += byte_r;
		} while (leftr > 0);
		sumbyte += tot;
	}
	D(msg(partition->device_name));
	D(msg_int("bytes read", sumbyte));

	return 0;
}

/* Return 0 if buffer can be filled to len by reading from datafd or -1. */
int
read_to_buffer(int datafd, char *buffer, int len)
{
	size_t leftr;
	int tot;
	ssize_t byte_r;
	enum read_code status;

	leftr = len;
	tot = 0;
	do {
		if (read_with_timeout(datafd, buffer + tot, leftr, &byte_r) < 0) {
			msg(error_fimage);
			return -1; /* timed out */
		}
		status = read_status(byte_r);
		if (status == READ_NO_DATA) {
			continue;
		} else if (status == END_OF_DATA) {
			msg(error_fimage);
			D(msg(__FUNCTION__));
			return -1;
		}

		leftr -= byte_r;
		tot += byte_r;
	} while (leftr > 0);

	return 0;
}

/* Return 0 if an end_token is found within 2 kByte by reading one byte
   sequentially or return -1. */
int
read_additional_data(int datafd, const char *end_token)
{
	size_t tokenlen;
	char buffer[2048];
	int pos;
	int len;

	D(msg(__FUNCTION__));

	tokenlen = strlen(end_token);

	if (read_to_buffer(datafd, buffer, 4) < 0) {
		return -1;
	}
	len = 4;
	pos = 0;
	while (strncmp(&buffer[pos], end_token, tokenlen) != 0 && len < 2048) {
		if (read_to_buffer(datafd, &buffer[len], 1) < 0) {
			return -1;
		}
		len++;
		pos++;
	}

	if (len < 2048) {
		D(msg("Found token"));
		return 0;
	} else {
		return -1;
	}
}

/* Read remaining data from datafd and ignore it. */
void
read_rest(int datafd)
{
	unsigned char buf[1024];
	ssize_t byte_r;

	D(msg(__FUNCTION__));

	if (datafd < 0) {
		return;
	}

	do {
		if (read_with_timeout(datafd, buf, 1024, &byte_r) < 0)
			return; /* timed out */
	} while (byte_r > 0);

}

static int
read_with_timeout(int fd, void *buf, size_t count, ssize_t *byte_r)
{
	fd_set rset;
	struct timeval tv;

	FD_ZERO(&rset);
	FD_SET(fd, &rset);

	tv.tv_sec = 300;
	tv.tv_usec = 0;

	if (select(fd + 1, &rset, NULL, NULL, &tv) <= 0) {
		return -1;
	}

	*byte_r = read(fd, buf, count);
	return 0;
}
